#include <fstream.h>
#include <hpldbin.h>
#include <dpldbin.h>
#include <popsmpl.h>
#include <slctsus.h>
#include <schemata.h>
#include <ga.h>
#include <Normal.h>
#include <GetOpt.h>

ParameterMap params(16,-1.28,1.28);
Normal  gauss_0_1(0,1,&mlcg); 

unsigned nGen = 400 , uPopSize = 30, uTest = 1;
float    flMRate = 0.003;
float    flXRate = 1;                              
float    flGap   = 0.1;                      
double   flHaploidProp = 1;
SigmaScaling<> *pScale = new SigmaScaling<>(1.5);
ofstream results("bog.dat");

//                             30
//                           ------
//      F4 : f4(xi) =        \       [ i * (xi)**4  + Gauss(0,1) ]
//                           /
//                           ------
//                             1
//
//              -1.28 <= xi <= 1.28 
//                                    
class F4                            
{
public:                      
        double operator()(const TGAObj * obj)
        {                                      
             double   flTemp, flTotal = 0;
             for(int i=0, j=0; i < 30; i++, j+=8) {
                 flTemp =  params((BinIndividual*) obj,j);
                 flTemp = i * pow(flTemp, 4.0);
                 flTemp += gauss_0_1();
                 flTotal += flTemp;
             }

             return flTotal;
        }
};

void    printBanner(const char *prog)
{
           cout << "usage: " << prog << " [-n gen] [-p popsize] [-x xpt] [-m mut] [-g gap] [-h percent]" << endl;
           cout << "-n gen      the number of generations to perform" << endl;
           cout << "-p popsize  the population size" << endl;
           cout << "-x xrate    the crossover rate" << endl;
           cout << "-m mrate    the mutation rate" << endl;
           cout << "-g gap      the generation gap" << endl;
           cout << "-h percent  the percentage of haploids in the population" << endl;
}

class HammingDistance 
{
public:
	double operator()(const TGAObj* obj1, const TGAObj* obj2) const
        {
		return ((const BinIndividual *) obj1)->hammingDistance(* (const BinIndividual *) obj2);
        }
};

void    runtest(double haploid_proportion)
{
        SimplePopulation<F4>* pPop = new SimplePopulation<F4>();
        int i;
        int hp = uPopSize * haploid_proportion;

        for (i=0; i<hp; i++)
             pPop->add(new BinHaploid(480,true));
        for (;i<uPopSize;i++)
             pPop->add(new BinDiploid(480,true));

        SelectionScheme<>* pSelect = new SUS_Select<vector<TGAObj*> >();
        Crossover* pXover = new BinTwoPtCrossover(flXRate); 
        GeneticAlgorithm<>  ga(pPop,
                              pSelect, pXover,                               
                              flMRate,                                         
                              flGap, pScale,true);
             
        cout << "Generation " << ga.trials() << endl;
        cout << "Best Ind" << endl << "-----------" << endl;
        cout << *ga.bestInd() << endl;
        cout << "Worst Ind" << endl << "-----------" << endl;
        cout << *ga.worstInd() << endl;

        while (ga.trials() < nGen) 
             ga.evolve();

        results << "Selection performed : " << ga.numSelected() << endl;
        results << "Crossover performed : " << ga.numCrossed() << endl;
        results << "Best Individual : " ;
        ga.bestInd()->printOn(results);
        results << endl;
        results << "Worst Individual : " ;
        ga.worstInd()->printOn(results);
        results << endl;
        results << "Online : " << ga.objOnLine() << endl;
        results << "Offline : " << ga.objOffLine() << endl << endl;
        results << "Final population" << endl;
        results << "----------------" << endl;
	ga.printOn(results);
        results << endl;
}

void    main(int argc, char *argv[])                   
{
        GetOpt options(argc,argv,"n:p:x:m:g:h:");
        char opt;
                    
        if (argc == 1) {   
           printBanner(argv[0]);                     
           exit(-1);
        }
        else {                                           
           opt = options();                             
           if (opt < 0 || opt == EOF) {
               printBanner(argv[0]);
               exit(-1);                                                       
           }
           else      
             do {
                 switch (opt) {
                 case 'n' :
                    nGen = abs(atoi(options.optarg));
                    break;
                 case 'p' :
                    uPopSize = atoi(options.optarg);
                    break;
                 case 'x' :
                    flXRate = fabs(atof(options.optarg));
                    if (flXRate > 1)
                        flXRate = 1;
                    break;
                 case 'm' :
                    flMRate = fabs(atof(options.optarg));
                    if (flMRate > 1)
                        flMRate = 1;
                    break;
                 case 'h' :
                    flHaploidProp = fabs(atof(options.optarg));
                    if (flHaploidProp > 1)
                        flHaploidProp = 1;
                    break;
                 case 'g' :
                    flGap = fabs(atof(options.optarg));
                    if (flGap > 1)
                        flGap = 1;
                    break;
                 case '?' :
                    printBanner(argv[0]);
                    exit(-1);
                 }
             } while ((opt = options()) != EOF);
        }

        results << "Haploid (%)     : " << flHaploidProp << endl;
        results << "Generations     : " << nGen << endl;
        results << "Population size : " << uPopSize << endl;
        results << "Crossover rate  : " << flXRate << endl;
        results << "Mutation rate   : " << flMRate << endl;
        results << "Generation gap  : " << flGap << endl << endl;

        runtest(flHaploidProp);
       
        cout << "Results saved in bog.dat" << endl;
}


