//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      Chinese University of Hong Kong
//
//      A GA program developed using TOLKIEN to solve
//      some multimodal functions by Mahfoud's deterministic crowding
//
//      Reference :
//
//      Mahfoud, S.W. (1992).  Crowding and Preselection Revisited.
//      In Parallel Problem Solving from Nature, 2.  Springer-Verlag. 1992.
//
//      This GA (or my implementation of it) works only for
//      function optimizations such that the fittest individual
//      has the largest objective value.
//      ( ref. DeterminsiticCrowding::generation() )
//
//
#include <BitString.h>
#include <iostream.h>
#include <ga.h>
#include <popsmpl.h>
#include <slcttour.h>
#include <mmfuncs.h>
#include <ranksb.h>
#include <GetOpt.h>          

SigmaScaling<> *pScale;
Crossover *pXvr = NULL;

class	DeterministicCrowding : public GeneticAlgorithm<>
{
public:
	DeterministicCrowding(PopulationBase<>*,
			      Crossover*,
                              ScalingScheme<>*,
                              double);

	void    evolve();

        const TGAObj* bestInd() const
        {          
            return pBest;
        }         

        const TGAObj* worstInd() const        
        {               
            return pWorst;
        }

protected:
	vector<int> mates;  // indices to individuals
        const TGAObj* pBest;
        const TGAObj* pWorst;
        findWorstInd() 
        {
            vector<TGAObj*>::const_iterator i = pPop->begin();
            pWorst = *i;
            i++;
            for (;i < pPop->end();i++)
                 if (pWorst->objValue() >= (*i)->objValue())
                     pWorst = *i;
        }
};

DeterministicCrowding::DeterministicCrowding(PopulationBase<>* pInds,
				   Crossover* pCrossover,
				   ScalingScheme<>* pScaleScheme,
				   double mutate_rate) :
                        mates(pInds->popSize(),0)
{
        flMRate = rangeCheck(mutate_rate,0.0,1.0, DEFAULT_MUTATRATE);
        pPop = pInds;
        pXover = pCrossover;
        pScale = pScaleScheme;
	reEvaluate = true;
        nTrials = 1;
        
        //                        
        //      ensure an even number of individuals    
        //                                                      
        if ((pPop->popSize() % 2) != 0) {  
            if (pPop->popSize() > 0)                   
                pPop->add( (* pPop->begin())->clone() );
                         
            if ((pPop->popSize() % 2) != 0) 
                    exit(-1);                                   
         }
                 

         statistics();
         pBest = pPop->bestInd();
         pWorst = pPop->worstInd();
         flOnLine = pPop->fitStat().mean();
         flOffLine = pPop->fitStat().max();
}

void DeterministicCrowding::evolve()
//
//	arrange the parents in a random manner into n/2 pairs
//
//      for each pair of individuals do
//              generate two offsprings,
//              the best two individuals among
//              the two parents and the two offsprings
//              are allowed to survive
//      enddo
//
{
	int i, j;
	TGAObj*  pChild1, *pChild2;
	double   diff1, diff2;
	int tmp;
        vector<TGAObj*>::iterator begin = pPop->begin(), p1, p2;
        bool findWorst = false;

	nTrials++;      // one more trial

	// create a random sequence
	for (i = 0; i < pPop->popSize(); i++)
	     mates[i] = i;
	for (i = pPop->popSize() - 1; i > 0; i--) {
		j = tRand(0,i);
		tmp = mates[j];
		mates[j] = mates[i];
		mates[i] = tmp;
	}

	for (i=0; i < pPop->popSize(); i+= 2) {
	     p1 = begin + mates[i];
	     p2 = begin + mates[i + 1];
	     pXover->setMates(*p1,*p2);
	     pChild1 = pXover->create();
	     pChild2 = pXover->create();
	     pChild1->mutate(flMRate);
	     pChild2->mutate(flMRate);
	     pPop->evaluate(pChild1);
	     pPop->evaluate(pChild2);

	     diff1 = abs((*p1)->objValue() - pChild1->objValue()) +
		     abs((*p2)->objValue() - pChild2->objValue());
	     diff2 = abs((*p1)->objValue() - pChild2->objValue()) +
		     abs((*p2)->objValue() - pChild1->objValue());

	     if (diff2 > diff1) {
		// parent1 <-> child1, parent2 <-> child2
		if ((*p1)->objValue() < pChild1->objValue()) {
                    if (*p1 == pBest)
                        pBest = pChild1;
                    else 
                        if (*p1 == pWorst)
                            findWorst = true;
                    delete *p1;
		    *p1 = pChild1;
                }
		else
		    delete pChild1;

		if ((*p2)->objValue() < pChild2->objValue()) {
                    if (*p2 == pBest)
                        pBest = pChild2;
                    else 
                        if (*p2 == pWorst)
                            findWorst = true;
                    delete *p2;
		    *p2 = pChild2;
                }
		else
		    delete pChild2;
	     }
	     else {
		// parent1 <-> child2, parent2 <-> child1
		if ((*p2)->objValue() < pChild1->objValue()) {
                    if (*p2 == pBest)
                        pBest = pChild1;
                    else 
                        if (*p2 == pWorst)
                            findWorst = true;
                    delete *p2;
		    *p2 = pChild1;
                }
		else
		    delete pChild1;

		if ((*p1)->objValue() < pChild2->objValue()) {
                    if (*p1 == pBest)
                        pBest = pChild2;
                    else 
                        if (*p1 == pWorst)
                            findWorst = true;
                    delete *p1;
		    *p1 = pChild2;
                }
		else
		    delete pChild2;
	     }
	}

        if (findWorst)
            findWorstInd();

	statistics();
}

unsigned nGen = 100 , uPopSize = 100, uTest = 1;
float    flMRate = 0.003;
float    flXRate = 1;

void printBanner(const char* prog)
{
    cout << "usage : " << prog << " [-f func] [-n gen] [-p popsize] [-x xpt] [-m mut]" << endl;

    cout << "-n gen      the number of generations to perform" << endl;
    cout << "-p popsize  the population size" << endl;
    cout << "-x xrate    the crossover rate" << endl;
    cout << "-m mrate    the mutation rate" << endl << endl;       
    cout << "-f [1,7] where " << endl;
    cout << "    1 = two peaks" << endl;
    cout << "    2 = Himmelblau's function" << endl;
    cout << "    3 = even peaks" << endl;
    cout << "    4 = uneven peaks" << endl;
    cout << "    5 = uneven decreasing peaks" << endl;
    cout << "    6 = two peaks trap" << endl;
    cout << "    7 : uneven cos" << endl;
}

void main(int argc, char* argv[])
{
	int optfunc = 1, chromlen;
	int i;
        GetOpt options(argc,argv,"f:n:p:x:m:");
        char opt;
        PopulationBase<>* pPop;
                   
        if (argc == 1) {   
           printBanner(argv[0]);
           exit(-1);
        }
        else {
           opt = options();                          
           if (opt < 0 || opt == EOF) {
               printBanner(argv[0]);
               exit(-1);                                
           }
           else
             do {
                 switch (opt) {
                 case 'f' :
                    optfunc = atoi(options.optarg);
                    if ((uTest < 1) || (uTest > 7))
                        uTest = 1;
                    break;
                 case 'n' :
                    nGen = abs(atoi(options.optarg));
                    break;
                 case 'p' :
                    uPopSize = atoi(options.optarg);
                    break;
                 case 'x' :
                    flXRate = fabs(atof(options.optarg));
                    if (flXRate > 1)
                        flXRate = 1;
                    break;
                 case 'm' :
                    flMRate = fabs(atof(options.optarg));
                    if (flMRate > 1)
                        flMRate = 1;
                    break;
                 case '?' :
                    printBanner(argv[0]);
                    exit(-1);
                 }
             } while ((opt = options()) != EOF);
        }

        cout << "Generations     : " << nGen << endl;
        cout << "Population size : " << uPopSize << endl;   
        cout << "Crossover rate  : " << flXRate << endl;
        cout << "Mutation rate   : " << flMRate << endl;


	switch(optfunc) {
		case 2 :
		pScale = new SigmaScaling<>(2);
                pPop = new SimplePopulation<Hmb>();
		chromlen = 40;
		break;
		case 3 :
		pScale = NULL;
                pPop = new SimplePopulation<Evenpeaks>();
		chromlen = 30;
		break;
		case 4 :
		pScale = NULL;
                pPop = new SimplePopulation<Unevenpeaks>();
		chromlen = 30;
		break;
		case 5 :
		pScale = NULL;
                pPop = new SimplePopulation<Unevendecpeaks>();
		chromlen = 30;
		break;
		case 6 :
		pScale = NULL;
                pPop = new SimplePopulation<Twopeaktrap>();
		chromlen = 20;
		break;
		case 7 :
		pScale = NULL;
                pPop = new SimplePopulation<Unevencos>();
		chromlen = 30;
		break;

                default:
		pScale = NULL;
                pPop = new SimplePopulation<Twopeaks>();
		chromlen = 20;
		break;
	}
	
        pXvr = new EvenPtCrossover<BinIndividual>(2,flXRate);

        switch(optfunc) {
                case 2:
                for (i=0; i < uPopSize; i++)
                     pPop->add(new HmbInd());
                break;

                case 3:
                case 4:
                case 5:
		case 7:
                for (i=0; i < uPopSize; i++)
                     pPop->add(new SinInd());
                break;

                default:
                for (i=0; i < uPopSize; i++)
                     pPop->add(new BinHaploid(chromlen,true,true));
                break;
        }

        DeterministicCrowding ga(pPop,
				 pXvr,pScale,
                                 flMRate);

        cout << "Generation " << ga.trials() << endl;
        cout << "Best Ind" << endl << "-----------" << endl;
        cout << *ga.bestInd() << endl;
        cout << "Worst Ind" << endl << "-----------" << endl;
        cout << *ga.worstInd() << endl;
        
        while (ga.trials() < nGen) {   
             ga.evolve();
/*                                                   
             cout << "Generation " << ga.trials() << endl;
             if (ga.trials() % 50 == 0)
                 ga.resetMonitor();

             cout << "Best Ind   : ";
             cout << *ga.bestInd() << endl;

             cout << "Worst Ind  : ";
             cout << *ga.worstInd() << endl;
*/
        }

        cout << "Final population" << endl;
        cout << "----------------" << endl;
        ga.printOn(cout);
        cout << endl;
        cout << "Crossover performed : " << ga.numCrossed() << endl;
        cout << "Best Individual : " ; 
        ga.bestInd()->printOn(cout);
        cout << endl;                                
        cout << "Worst Individual : " ;
        ga.worstInd()->printOn(cout);  
        cout << endl;
        cout << "Online : " << ga.onLine() << endl;
        cout << "Offline : " << ga.offLine() << endl;
}

