// $Header: /home/anthony/tolkien/include/RCS/chromo.h,v 1.4 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined( __CHROMO_H )
#define __CHROMO_H

#include <iostream.h>
#include <tgaobj.h>
#include <ranksb.h>
#include <algo.h>
#include <xover.h>

template <class Gene>
void evenPtCrossover(const vector<Gene>&  mother, 
                     const vector<Gene>&  father,
                     vector<Gene>& child,
                     int nXpts) 
{
     int i, j = nXpts - 1, len = mother.size() - 1, cont, k;
     vector<Gene>::iterator begin2, first2;
     vector<Gene>::const_iterator begin1, first1, last1;
     vector<int> xpts;

     child = mother;

     do {
        cont = 0;
        // since we treat the chromosome as a ring
        // 0 is equivalent to (size() - 1)
        // ensure the first point and the last point
        // do not overlap

        ranksb.ordered(xpts,nXpts,mother.size());

        for (i=nXpts-1; i > 0 && cont == 0; i--)
             if (xpts[i] == (xpts[i-1]+1))
                 cont = 1;
             
        if (xpts[0] == 0 && xpts[j] == len)
            cont = 1;
     } while ( cont == 1 );

     begin1 = father.begin();
     begin2 = child.begin();
     for (i = 0; i <= j; i+= 2) {
          first1 = begin1 + xpts[i];
          first2 = begin2 + xpts[i];
          last1  = first1 + (xpts[i+1] - xpts[i] + 1);
          copy(first1, last1, first2);
     }
}

template <class Gene>
void oddPtCrossover(const vector<Gene>&  mother, 
                    const vector<Gene>&  father,
                    vector<Gene>& child,
                    int nXpts) 
{
     vector<int> xpts;
     vector<Gene>::iterator begin2, first2;
     vector<Gene>::const_iterator begin1, first1, last1;
     int i, j, p, q;

     child = mother;
     ranksb.ordered(xpts,nXpts,father.size() - 1);
     // generate an ordered sequence in the range [1,size - 2]
     // in order to preserve the two segments 
     // [0 - n] and [m - ( size -1 )]
     // n is the first value in the sequence 
     // m is the last value in the sequence

     begin1 = father.begin();
     begin2 = child.begin();
     for (i=0, j = nXpts - 1; i < j; i+=2) {
          p = xpts[i] + 1;
          q = xpts[i+1] + 1;
          first1 = begin1 + p;
          first2 = begin2 + p;
          last1  = first1 + (q - p);
          copy(first1, last1, first2);
     }
     // copy the last segment
     p = xpts[nXpts - 1] + 1;
     first1 = begin1 + p;
     first2 = begin2 + p;
     last1  = father.end();
     copy(first1, last1, first2);
}

template <class Gene>
void uniformCrossover(const vector<Gene>&  mother, 
                      const vector<Gene>&  father,
                      vector<Gene>& child,
                      double flXRate)
{
     vector<Gene>::iterator begin1, last1, cur1;
     vector<Gene>::const_iterator cur2;

     child = mother;
     begin1 = cur1 = child.begin();
     last1  = child.end();
     cur2   = father.begin();

     for (; cur1 != last1; cur1++, cur2++) 
          if ( flip(flXRate) )
               *cur1 = *cur2;
}

template <class IndClass>
class OrderCrossover : public Crossover
{
public:
      OrderCrossover(double flRate=DEFAULT_XOVERATE) :
          numCreated(0), pFather(NULL), pMother(NULL)
      {
           rate(flRate);
      }

      TGAObj*  create()
      {
           const IndClass* pTmp = (const IndClass*) pFather;
           TGAObj*  pChild;

           if (flip())
              pChild = ((const IndClass*) pMother)->
                       orderCrossover((const IndClass*) pFather);
           else
              pChild = pMother->clone();

           // swap parents
           pFather = pMother;
           pMother = pTmp;

           count();
           return pChild;
      }
        
      Crossover* clone() const
      {
           return new OrderCrossover<IndClass>(*this);
      }

      double rate() const
      {
          return flXRate;                  
      }
        
      void rate(double flNewRate)    
      {
          flXRate = rangeCheck(flNewRate,0.0,1.0,DEFAULT_XOVERATE);
      }

      long total() const      
      {
          return numCreated;                           
      }

      void setMates(const TGAObj * ind1, const TGAObj * ind2)
      {
          pMother = ind1;
          pFather = ind2;               
      }

      void getMates(const TGAObj * & pInd1, const TGAObj * & pInd2) const
      {
          pInd1 = pMother;
          pInd2 = pFather;
      }

protected:
      double flXRate;
      long numCreated;
      const TGAObj *  pMother;
      const TGAObj *  pFather;
        
      void count() { numCreated++; }       
};

template <class Gene>
void orderCrossover(const vector<Gene>& mother,
                    const vector<Gene>& father,
                    vector<Gene>& child)
{
    vector<Gene>::iterator i, j, k;
    int from, to;

    child = mother;

    do {
        from = tRand.asLong(1, mother.length() - 1);
        to = tRand.asLong(1, mother.length() - 1);
    } while (from == to);

    if (from > to) 
        swap(from,to);

    // the segment [from,to) from mother is kept in the child
    vector<Gene> segment;
    segment.insert(mother.begin() + from, mother.begin() + to);
    sort(segment.begin(),segment.end());

    i = j = father.begin();
    i += from;
    j += to;

    k = child.begin() + to;

    while (i != j) {
       if (j == father.end())
           j = father.begin();
       if (k == child.end())
           k = child.begin();
       if (! binary_search(segment.begin(),segment.end(),*j)) {
           *k = *j;
           k++;
       }
    }
}

#endif

