// $Header: /home/anthony/tolkien/include/RCS/ga.h,v 1.8 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined( __GA_H )
#define __GA_H

#include <populatn.h>
#include <xover.h>
#include <slctrw.h>
#include <vector.h>

template <class Collection = vector<TGAObj*> >
class   GeneticAlgorithm 
{
public:
	GeneticAlgorithm(PopulationBase<Collection> *pInds,
			 SelectionScheme<Collection> *pSelectScheme,
			 Crossover *pCrossover,
			 double mutate_rate,
			 double genGap,
			 ScalingScheme<Collection> *pScaleScheme = NULL,
			 bool reEval = false) :
                         flMRate(rangeCheck(mutate_rate,0.0,1.0,
                                            DEFAULT_MUTATRATE)),
                                   pPop(pInds),
                                   pXover(pCrossover),
                                   pSSch(pSelectScheme),
                                   pScale(pScaleScheme),
                                   nTrials(1),
                                   flOnLine(0),
                                   flOffLine(0),
                                   reEvaluate(reEval)
        {
            double flGenerationGap = rangeCheck(genGap,0.0,1.0,1.0);

            //                                                     
            //      ensure an even number of individuals     
            //                                                             
            if ((pPop->popSize() % 2) != 0) {                                  
                if (pPop->popSize() > 0)            
                    pPop->add( (* pPop->begin())->clone() );
                                   
                if ((pPop->popSize() % 2) != 0) {            
                    exit(-1);                               
                }
            }
                                   
            if ((nOffSprings = (int) (pPop->popSize() * flGenerationGap)) < 2)
                 nOffSprings = 2;  // minimum number of new offsprings is two

            pSSch->selectOn(*pPop);
            statistics();
            flOnLine = pPop->fitStat().mean();
            flOffLine = pPop->fitStat().max();
            flObjOnLine = pPop->objStat().mean();
            flObjOffLine = pPop->objStat().max();
        }

	virtual ~GeneticAlgorithm()
        {
            if (pSSch)
                delete pSSch;
            if (pXover)
                delete pXover;
            if (pPop)
                delete pPop;
            if (pScale)
                delete pScale;
        }

        virtual void   printOn(ostream& out) const
        {
               pPop->printOn(out);
               out << endl;
        }

	virtual void evolve()
        {
              nTrials++;      // one more trial
              pSSch->init();
              pPop->evolve(pXover,pSSch,pScale,nOffSprings,flMRate,reEvaluate);
              statistics();
        }

        void    gap(double flNewGap)
        {
            flNewGap = rangeCheck(flNewGap,0.0,1.0,1.0);
            if ((nOffSprings = (int) (pPop->popSize() * flNewGap)) < 2)
                 nOffSprings = 2;  // minimum number of new offsprings is two
        }

	virtual void statistics()
        {
              flOnLine = 
                 (flOnLine*(nTrials - 1)+pPop->fitStat().mean()) / nTrials;
              flOffLine =
                 (flOffLine*(nTrials - 1)+pPop->fitStat().max()) / nTrials;
              flObjOnLine = 
                 (flObjOnLine*(nTrials - 1)+pPop->objStat().mean()) / nTrials;
              flObjOffLine =
                 (flObjOffLine*(nTrials - 1)+pPop->objStat().max()) / nTrials;
        }

	void  selectionScheme(SelectionScheme<Collection>* pNewSch)
        {
	      delete pSSch;
	      pSSch = pNewSch;
              pSSch->selectOn(*pPop);
        }

	void  crossover(Crossover* pNewXOver)
        {
	      delete pXover;
	      pXover = pNewXOver;
        }

	void  scaleScheme(ScalingScheme<Collection> *pNewScale)
        {
            delete pScale;
            pScale = pNewScale;
        }    

	virtual int trials() const
        {
	    return nTrials;
        }

	virtual const TGAObj* bestInd() const
        {
	    return    pPop->bestInd();
        }

	virtual const TGAObj* worstInd() const
        {
	    return    pPop->worstInd();
        }

        virtual long numCrossed() const
        {
	    return pXover->total();
        }

        virtual long numSelected() const
        {
	    return pSSch->total();
        }

        virtual double onLine() const
        {
	    return flOnLine;
        }

        virtual double offLine() const
        {
	    return flOffLine;
        }

        virtual double objOnLine() const
        {
	    return flObjOnLine;
        }

        virtual double objOffLine() const
        {
	    return flObjOffLine;
        }

protected:

	PopulationBase<Collection>* pPop;
	Crossover*  pXover;
	SelectionScheme<Collection>* pSSch;

	int nTrials;       // trials counter

	int nOffSprings;   // number of offsprings
			   // created in each generation

	double flMRate;    // mutation rate

	bool reEvaluate;   // whether to re-evaluate
			   // old individuals

	ScalingScheme<Collection>  *pScale;

	double	flOnLine;               // on-line performances
	double  flOffLine;              // off-line performances

	double	flObjOnLine;            // raw on-line performances
	double  flObjOffLine;           // raw off-line performances

	GeneticAlgorithm() : pPop(NULL), pXover(NULL), 
                             pSSch(NULL), pScale(NULL)
        {
        }
};

#endif

