// $Header: /home/anthony/tolkien/include/RCS/genitor.h,v 1.4 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
//      Reference material :
//      ICGA-91 : The GENITOR Algorithm and Selection Pressure :
//                Why Rank-Based Allocation of Reproductive Trials is Best
//
//
#if !defined( __GENITOR_H )
#define __GENITOR_H

#include <ga.h>
#include <algo.h>

template <class PhenotypeFunction>
class GENITOR : public GeneticAlgorithm<> 
{
public:
	GENITOR(TGAObj* pInd, int nPopSize,
		Crossover *pCrossover,
		double mutate_rate, double bias = 1.5) : 
        GeneticAlgorithm<>(), nSelected(0)
        {
            int i;
            double val;
            SampleStatistic  _objStat;

            flMRate=rangeCheck(mutate_rate,0.0,1.0, DEFAULT_MUTATRATE);
            pXover = pCrossover;
            nTrials = 1;

            flBias = bias > 1 ? bias : 1.5;
            flBiasSqr = flBias * flBias;
            flBiasLessOne = flBias - 1;

	    sorted.reserve(nPopSize);
            sorted.push_back(pInd);
            for (i=1;i<nPopSize;i++)
                 sorted.push_back(pInd->clone());
            for (i=0; i<nPopSize;i++) {
                 sorted[i]->randomize();
                 evaluate(sorted[i]);
                 _objStat += sorted[i]->fitness();
            }

            flMean = flOnLine = _objStat.mean();
            flOffLine = _objStat.max();
            flObjOnLine = _objStat.mean();
            flObjOffLine = _objStat.max();
            sort(sorted.begin(),sorted.end(),DecFitness());
        }

	GENITOR(vector<TGAObj*> inds,
		Crossover *pCrossover,
		double mutate_rate, double bias) : 
        GeneticAlgorithm<>(), nSelected(0)
        {
            int i;
            double val;
            SampleStatistic  _objStat;

            flMRate=rangeCheck(mutate_rate,0.0,1.0, DEFAULT_MUTATRATE);
            pXover = pCrossover;
            nTrials = 1;

            flBias = bias > 1 ? bias : 1.5;
            flBiasSqr = flBias * flBias;
            flBiasLessOne = flBias - 1;

            sorted = inds;
            for (i=sorted.size() - 1; i >= 0; i--) {
                 evaluate(sorted[i]);
                 _objStat += sorted[i]->fitness();
            }
            flMean = flOnLine = _objStat.mean();
            flOffLine = _objStat.max();
            flObjOnLine = _objStat.mean();
            flObjOffLine = _objStat.max();
            sort(sorted.begin(),sorted.end(),DecFitness());
        }

        void printOn(ostream& out) const
        {
             for (vector<TGAObj*>::const_iterator i = sorted.begin();
                  i != sorted.end(); ++i) {
                  (*i)->printOn(out);
                  out << endl;
             }     
        }

	void evolve()
        {
             TGAObj  *pInd;
             double  oldWorst = sorted.back()->objValue();

             nTrials++;      // one more trial
             
             pXover->setMates(select(),select());
             pInd = pXover->create();
             pInd->mutate(flMRate);
             evaluate(pInd);
             sorted.pop_back();
             sorted.insert(lower_bound(sorted.begin(),sorted.end(),
                                       pInd,DecFitness()), pInd);
             flMean = (flMean * sorted.size() - 
                       oldWorst + pInd->fitness()) / sorted.size();

             statistics();
        }

        long numSelected() const
        {
             return nSelected;
        }

	void statistics()
        {
              flOnLine = 
                 (flOnLine*(nTrials - 1)+flMean) / nTrials;
              flOffLine =
                 (flOffLine*(nTrials - 1)+sorted.front()->fitness()) / nTrials;
              flObjOnLine = 
                 (flObjOnLine*(nTrials - 1)+flMean) / nTrials;
              flObjOffLine =
                 (flObjOffLine*(nTrials - 1)+sorted.front()->objValue()) / nTrials;
        }

	const TGAObj* bestInd() const
        {
	    return sorted.front();
        }

	const TGAObj* worstInd() const
        {
	    return sorted.back();
        }


protected:
	vector<TGAObj*> sorted;
        PhenotypeFunction func;
        double flBias, flBiasSqr, flBiasLessOne;
        int nSelected;
        double flMean;

        const TGAObj* select() 
        {
              int index;

              index = (int) (sorted.size() *
                 (flBias -
                 sqrt(flBiasSqr - tRand.asDouble() * 4 * flBiasLessOne)) /
                 (2 * flBiasLessOne));

              nSelected++;

              return sorted[index];
        }

        void evaluate(TGAObj* pObj)
        {
               double val = func(pObj);
               pObj->objValue(val);
               pObj->fitness(val);
        }

};

#endif

