// $Header: /home/anthony/tolkien/include/RCS/haploid.h,v 1.6 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//
#if !defined( __HAPLOID_H )

#define __HAPLOID_H

#include <tgaobj.h>
#include <vector.h>
#include <ranksb.h>
#include <iostream.h>
#include <algo.h>
#include <math.h>

template <class Gene>
class Haploid : public TGAObj
{
public:
    typedef vector<Gene>::const_iterator const_iterator;
    typedef vector<Gene>::iterator iterator;

    Haploid() { }

    Haploid(int n, const Gene& x) : genes(n,x) { }

    Haploid(const Haploid<Gene>& src) : genes(src.genes) { }

    Haploid(const vector<Gene>& src) : genes(src) { }

    const_iterator begin() const { return genes.begin(); }
    iterator begin() { return genes.begin(); }

    const_iterator end() const { return genes.end(); }
    iterator end() { return genes.end(); }

    Haploid<Gene>*  evenPtCrossover(const Haploid<Gene>* mate, int nXpts) const
    {
        Haploid<Gene> *pMate = (Haploid<Gene>* ) mate, 
                      *pChild = (Haploid<Gene>* ) clone();
        int i, j = nXpts - 1, len = length() - 1, cont, k;
        iterator begin2, first2;
        const_iterator begin1, first1, last1;
        vector<int> xpts;

        do {
           cont = 0;
           // since we treat the chromosome as a ring
           // 0 is equivalent to (length() - 1)
           // ensure the first point and the last point
           // do not overlap

           ranksb.ordered(xpts,nXpts,length());
           for (i=nXpts-1; i > 0 && cont == 0; i--)
             if (xpts[i] == (xpts[i-1]+1))
                 cont = 1;

           if (xpts[0] == 0 && xpts[j] == len)             
               cont = 1;
        } while ( cont == 1 );

        begin1 = pMate->begin();
        begin2 = pChild->begin();
        for (i = 0; i <= j; i+= 2) {
          first1 = begin1 + xpts[i];
          first2 = begin2 + xpts[i];
          last1  = first1 + (xpts[i+1] - xpts[i] + 1);
          copy(first1, last1, first2);
        }

        return pChild;
    }

    Haploid<Gene>*  oddPtCrossover(const Haploid<Gene>* mate, int nXpts) const
    {
        Haploid<Gene> *pMate = (Haploid<Gene>* ) mate, 
                      *pChild = (Haploid<Gene>* ) clone();
        vector<int> xpts;
        vector<Gene>::iterator begin2, first2;
        vector<Gene>::const_iterator begin1, first1, last1;
        int i, j, p, q;

        ranksb.ordered(xpts,nXpts,length() - 1);
        // generate an ordered sequence in the range [1,length - 2]
        // in order to preserve the two segments
        // [0 - n] and [m - ( length -1 )]
        // n is the first value in the sequence
        // m is the last value in the sequence

        begin1 = pMate->begin();
        begin2 = pChild->begin();
        for (i=0, j = nXpts - 1; i < j; i+=2) {
             p = xpts[i] + 1;
             q = xpts[i+1] + 1;
             first1 = begin1 + p;
             first2 = begin2 + p;
             last1  = first1 + (q - p);
             copy(first1, last1, first2);
        }

        return pChild;
    }

    Haploid<Gene>* uniformCrossover(const Haploid<Gene> *pMate, double flXRate) const 
    {
           vector<Gene>::iterator begin1, last1, cur1, cur2;
           Haploid<Gene> *pChild = (Haploid<Gene>*) clone();

           begin1 = cur1 = pChild->begin();
           last1  = pChild->end();
           cur2   = ((Haploid<Gene>*) pMate)->begin();

           for (; cur1 != last1; cur1++, cur2++) 
                if ( flip(flXRate) )
                     *cur1 = *cur2;
           return pChild;
    }

    
    Haploid<Gene>* orderCrossover(const Haploid<Gene> *pMate) const 
    {
          vector<Gene>::const_iterator i, k;
          vector<Gene>::iterator j;
          int from, to;
          Haploid<Gene>* pChild = (Haploid<Gene>*) clone();

          do {
              from = tRand.asLong(1, length() - 1);
              to = tRand.asLong(1, length() - 1);
          } while (from == to);                                      
          
          if (from > to)                                           
              swap(from,to);

          // the segment [from,to) from mother is kept in the child
          vector<Gene> segment(begin() + from, begin() + to);
          sort(segment.begin(),segment.end());

          i = j = pChild->begin();
          i += from;
          j += to;

          k = pMate->begin();
          k += to;

          while (i != j) {
             if (! binary_search(segment.begin(),segment.end(),*k) ) {
                 *j = *k;
                 j++;
                 if (j == pChild->end())
                     j = pChild->begin();
             }
             k++;
             if (k == pMate->end())
                 k = pMate->begin();
          }

          return pChild;
    }

    int   length() const
    {
          return genes.size();
    }

    void printOn(ostream& out) const
    {
         for (const_iterator i = begin(); i != end(); i++) 
             out << *i;
	 out << ' ';
	 TGAObj::printOn(out);
    }

    double distance(const TGAObj *obj) const
    {
	double val = 0;
        const_iterator i = begin(), 
                       j = ((const Haploid<Gene>*) obj)->begin(),
                       last1  = end(), 
                       last2  = ((const Haploid<Gene>*) obj)->end();
        for (;i != last1 && j != last2; i++, j++)
             val += (*i - *j);
	return fabs(val);
    }

protected:
	vector<Gene> genes;
};

#endif

