// $Header: /home/anthony/tolkien/include/RCS/populatn.h,v 1.9 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
#if !defined( __POPULATN_H )
#define     __POPULATN_H

#include <tgaobj.h>
#include <iterator.h>
#include <vector.h>
#include <xover.h>
#include <SmplStat.h>
#include <math.h>
#include <select.h>
#include <scale.h>

template <class Collection = vector<TGAObj*> >
class PopulationBase
{
public:
        typedef Collection::const_iterator const_iterator; 
        typedef Collection::iterator iterator; 

        virtual ~PopulationBase()
        {
	   removeAll();
	}

        void removeAll()
        {
             for (iterator i=inds.begin(); i != inds.end(); i++)
                  delete *i;
             inds.erase(inds.begin(), inds.end());
        }

	long    evals() const
        {
	        return lEvals;
        }

	int	popSize() const 
        {
		return inds.size();
        }

        const_iterator begin() const
        {
                return inds.begin();
        }

        iterator begin() 
        {
                return inds.begin();
        }

        const_iterator end() const
        {
                return inds.end();
        }

        iterator end() 
        {
                return inds.end();
        }

	virtual void evaluate(TGAObj * pInd) = 0;

        virtual void add(TGAObj* pObj, bool bEval = true)
        {
             if (bEval)
                evaluate(pObj);
             inds.push_back(pObj);
             
             _objStat += pObj->objValue();
             _fitStat += pObj->fitness();
             if (pBestInd == NULL) 
                 pBestInd = pWorstInd = pObj;
             else {
                if ( pObj->fitness() >= pBestInd->fitness() )
                     pBestInd = pObj;
                else
                     if ( pObj->fitness() < pWorstInd->fitness() )
                         pWorstInd = pObj;
             }
        }

	virtual void printOn( ostream& out ) const
        {
           for (Collection::const_iterator i = inds.begin();
                i != inds.end(); ++i) {
                (*i)->printOn(out);
                out << endl;
           }
        }

	virtual void randomize()
	{
           for (Collection::iterator i = inds.begin();
                i != inds.end(); ++i) 
                (*i)->randomize();
	}

	virtual void mutate(double flRate)
	{
           for (Collection::iterator i = inds.begin();
                i != inds.end(); ++i) 
                (*i)->mutate(flRate);
        }

        virtual void evolve(Crossover* pXover,
                            SelectionScheme<Collection>* pSSch,
                            ScalingScheme<Collection>* pScale,
                            int nOffSprings,
                            double flMRate,
                            bool reEvaluate)
        {
              int i, j;
              vector<TGAObj*> newInds;                  
              TGAObj*  pTmpInd;
            
              newInds.reserve(nOffSprings);                                     
              i = nOffSprings;                             
              while (i > 0) {           
                  pXover->setMates(pSSch->select(), pSSch->select());           
                  for (j=2; j > 0 && i > 0; j--, i--) { 
                       pTmpInd = pXover->create();
                       pTmpInd->mutate(flMRate);
                       newInds.push_back(pTmpInd);
                  }                        
              }
               
              replace(newInds, reEvaluate);
              if (pScale)
                  (*pScale)(*this, pBestInd, pWorstInd);
              statistics();
        }

	virtual void statistics()
        {
           Collection::iterator i = inds.begin();
           _objStat.reset();
           _fitStat.reset();

           pBestInd = pWorstInd = *i;
           ++i;
           
           while( i != inds.end() ) {
                _objStat += (*i)->objValue();
                _fitStat += (*i)->fitness();
                if ( (*i)->fitness() > pBestInd->fitness() )
                     pBestInd = *i;
                else
                     if ( (*i)->fitness() < pWorstInd->fitness() )
                         pWorstInd = *i;
                ++i;
           }
        }

	SampleStatistic& objStat() 
        {
		return _objStat;
        }

	SampleStatistic& fitStat() 
        {
		return _fitStat;
        }

        const TGAObj* bestInd() const
        {
            return    pBestInd;
        }
        
        const TGAObj* worstInd() const
        {
            return    pWorstInd;                  
        }

	void resetStatistics()
        {
           _objStat.reset();
           _fitStat.reset();
           pBestInd = pWorstInd = NULL;
        }

protected:

        Collection inds;

        long  lEvals;   // number of evaluations performed

        SampleStatistic  _objStat;
        SampleStatistic  _fitStat;
                    
        TGAObj *pBestInd, *pWorstInd;        
                
        PopulationBase() : pBestInd(NULL), pWorstInd(NULL)
        { 
        }
        
        PopulationBase(const PopulationBase& src) : 
                   pBestInd(NULL),
                   pWorstInd(NULL)
        {
           inds.reserve(src.popSize());
           for (Collection::const_reverse_iterator i = src.inds.rbegin();
                i != src.inds.rend(); ++i) 
                add((*i)->clone(), false);
           statistics();
        }

	void addAll(const Collection& cltn)
        {
           inds.reserve(cltn.size());
           for (Collection::const_reverse_iterator i = cltn.rbegin();
                i != cltn.rend(); ++i) 
                add((*i)->clone(), false);
           statistics();
        }

        virtual void replace(vector<TGAObj*>& inds, bool eval = false) = 0;
};

template <class PhenotypeFunction, class Collection = vector<TGAObj*> >
class  Population : public PopulationBase<Collection>
{
public:

	void evaluate(TGAObj * pInd) 
	{
	     double val = func((const TGAObj *) pInd);
             pInd->objValue(val);
             pInd->fitness(val);
	}

protected:

        PhenotypeFunction func;
};

#endif

