// $Header: /home/anthony/tolkien/include/RCS/scale.h,v 1.5 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
#if !defined( __SCALE_H )
#define __SCALE_H

template <class Collection = vector<TGAObj*> > class PopulationBase;

template < class Collection = vector<TGAObj*> >
class ScalingScheme 
{
public:
	virtual ~ScalingScheme() { }

        virtual void  operator()(PopulationBase<Collection>& pop,
                                 TGAObj*& pBest, TGAObj*& pWorst)
        //
        //      scale by assigning random fitness to individuals
        //      this function is useful when the derived operator()
        //      scaled all fitnesses to the same value (or zero)
        //      this is possible at least for SigmaScaling
        //
        {
                Collection::iterator popIter = pop.begin();
                TGAObj* pObj;
                vector<int> ints;
                int i = 0;

                ranksb.random(ints, pop.popSize(), pop.popSize());

                // the fitness values are in the set [1 ... popSize()]
                pObj = *popIter;
                pObj->fitness( 1 + ints[i] );
                pWorst = pBest = pObj;
                ++popIter;
                while ( popIter != pop.end() ) {
                    pObj = *popIter;
                    pObj->fitness( 1 + ints[i] );
                    if (pObj->fitness() < pWorst->fitness())
                           pWorst = pObj;
                       else
                           if (pObj->fitness() > pBest->fitness())
                               pBest = pObj;
                       ++popIter;
                }
        }
};

template < class Collection = vector<TGAObj*> >
class	SigmaScaling : public ScalingScheme<Collection>
//                     _        
//      f(x) = g(x) - (g - c * sd )
//
//           f(x) is the fitness value
//           g(x) is the objective value (g(x) is the smaller the better)
//           _
//           g    is the mean of g(x)
//           sd    is the SD of g(x)
//           c    is a value between 0 - 3
{
public:
	SigmaScaling(double flValue)
        {
	    scaleFactor(flValue);
        }

	virtual void scaleFactor(double flScaleFactor)
        {
	    double flOldValue = flSFactor;
	    flSFactor = flScaleFactor > 0 ?
		    ((flScaleFactor <= 3.0) ? flScaleFactor : 1.0 ) : 0;
        }

	double scaleFactor() const
        {
               return flSFactor;
        }

        void  operator()(PopulationBase<Collection>& pop,
                         TGAObj*& pBestInd, TGAObj*& pWorstInd)
        {
            Collection::iterator popIter = pop.begin();
            TGAObj* pObj;
               
            if (popIter != pop.end()) {
                double flDiff = (pop.objStat()).mean() -
                 flSFactor * (pop.objStat()).stdDev();
                pBestInd = pWorstInd = *popIter;
                (*popIter)->fitness( (*popIter)->objValue() - flDiff );
                if ((*popIter)->fitness() < 0)
                    (*popIter)->fitness( 0 );
                ++popIter;
                while ( popIter != pop.end() ) {
                    pObj = *popIter;
                    pObj->fitness( pObj->objValue() - flDiff );
                    if (pObj->fitness() < 0) {        
                        pObj->fitness( 0 );
                        pWorstInd = pObj;
                    }                                              
                    else if (pObj->fitness() > pBestInd->fitness())
                             pBestInd = pObj;
                    else if (pObj->fitness() < pWorstInd->fitness())      
                             pWorstInd = pObj;
                    ++popIter;
                }                                 
            }
            if (pBestInd->fitness() == pWorstInd->fitness())
                ScalingScheme<Collection>::operator()(pop,pBestInd,pWorstInd);
        }

protected:
	double flSFactor;
};

template < class Collection = vector<TGAObj*> >
class	LinearScaling : public ScalingScheme<Collection>
{
public:
       void  operator()(PopulationBase<Collection>& pop,
                         TGAObj*& pBest, TGAObj*& pWorst) {
           double flDelta, flA, flB, flAvg;                        
                       
           if (pop.objStat().min() >= 0) {
              flAvg = (pop.objStat().max() + pop.objStat().min()) / 2;
              flDelta = pop.objStat().max() - pop.objStat().min();
              flA = flAvg / flDelta;
              flB = flAvg * ( pop.objStat().max() - 2 * flAvg) / flDelta;
           }
           else {
              flDelta = flAvg - pop.objStat().min();
              flA = flAvg / flDelta;
              flB = pop.objStat().min() * -1 * flAvg / flDelta;
           }

           Collection::iterator popIter = pop.begin();
           while ( popIter != pop.end() ) {
                   (*popIter)->fitness( flA * (*popIter)->objValue() + flB );
                  ++popIter;
           }
       }
};

//
//                        k
//  Power scaling : f' = f 
//
template < class Collection = vector<TGAObj*> >
class	PowerScaling : public ScalingScheme<Collection>
{
public:
	PowerScaling(double flValue = 1.005) : flPower(flValue)
        {
            if (flPower < 0) 
                flPower = 1.005;
        }

        void  operator()(PopulationBase<Collection>& pop,
                         TGAObj*& pBestInd, TGAObj*& pWorstInd)
        {
            Collection::iterator popIter = pop.begin();
            TGAObj* pObj;
           
            if (popIter != pop.end()) {                               
                pBestInd = pWorstInd = *popIter;
                (*popIter)->fitness( pow((*popIter)->objValue(), flPower) );
                ++popIter;
                while ( popIter != pop.end() ) {               
                    pObj = *popIter;
                    pObj->fitness( pow(pObj->objValue(), flPower) );
                    
                    if (pObj->fitness() > pBestInd->fitness())
                        pBestInd = pObj;
                    else if (pObj->fitness() < pWorstInd->fitness())
                             pWorstInd = pObj;
                    ++popIter;
                }   
            }
            if (pBestInd->fitness() == pWorstInd->fitness())
                ScalingScheme<Collection>::operator()(pop,pBestInd,pWorstInd);
        }

protected:
	double flPower;
};

#endif

